#include "access_map.h"

/* Return the flexpage supporting 'position'. */

Flexpage *AccessMap::find(offset_t position)
{
    std::lock_guard<std::mutex> guard(_lock);

    _AccessMap::iterator it = _flexpages.upper_bound(position);

    if ((_flexpages.size() > 0) && (it != _flexpages.begin()))
    {
        it--;

        if (it->second->supports_position(position))
            return it->second;
    }

    return NULL;
}

/* Insert a mapping for 'flexpage'. */

void AccessMap::insert(Flexpage *flexpage)
{
    std::lock_guard<std::mutex> guard(_lock);

    _flexpages.insert(_AccessMapEntry(flexpage->base_offset, flexpage));
}

/* Remove the mapping supported by 'flexpage'.

   The flexpage may have obtained by another mapper before being purged from
   this object's mapping and before being purged from the queue (and thus
   disassociated from this mapper), leaving an opportunity for another mapper to
   now be removing it here. In such a situation, flushing has already occurred
   and will not be performed again. */

bool AccessMap::remove(PageOwner *owner, Flexpage *flexpage)
{
    std::lock_guard<std::mutex> guard(_lock);

    _AccessMap::iterator it = _flexpages.find(flexpage->base_offset);

    if (it != _flexpages.end())
    {
        owner->flush(flexpage, true);
        _flexpages.erase(flexpage->base_offset);
        return true;
    }

    return false;
}

/* Purge all flexpages, using the 'owner' to flush their contents and
   'pages' to make the flexpages available to other accessors. */

void AccessMap::purge(PageOwner *owner, PageCollection *pages)
{
    std::lock_guard<std::mutex> guard(_lock);

    _AccessMap::iterator it = _flexpages.begin(), entry;

    while (it != _flexpages.end())
    {
        entry = it;
        it++;

        Flexpage *flexpage = entry->second;

        /* Some flexpages may be unavailable in the queue. Only those
           that can be reserved should be flushed and made available
           again. */

        if (pages->reserve(owner, flexpage))
        {
            owner->flush(flexpage, true);
            pages->release(flexpage);
            _flexpages.erase(entry);
        }
    }
}

/* Flush flexpages in the given range from 'start' with 'size', using 'owner'
   and 'pages'. */

void AccessMap::flush_all(offset_t start, offset_t size, PageOwner *owner, PageCollection *pages)
{
    offset_t end = start + size;

    std::lock_guard<std::mutex> guard(_lock);

    /* The start may be within an existing flexpage where the flexpage size is a
       page multiple. */

    _AccessMap::iterator it = _flexpages.upper_bound(start), entry;

    if ((_flexpages.size() > 0) && (it != _flexpages.begin()))
        it--;

    /* Inspect flexpages at or after start until end. */

    while (it != _flexpages.end())
    {
        entry = it;
        it++;

        Flexpage *flexpage = entry->second;

        if (flexpage->base_offset >= end)
            break;

        /* Attempt to flush each flexpage, releasing ones that are no longer
           needed. */

        if (pages->reserve(owner, flexpage))
        {
            owner->flush(flexpage, false);

            /* Where no users of the flexpage persist, release the flexpage for
               reuse and remove this entry. */

            if (!flexpage->valid())
            {
                pages->release(flexpage);
                _flexpages.erase(entry);
            }
            else
                pages->queue(owner, flexpage);
        }
    }
}

// vim: tabstop=4 expandtab shiftwidth=4
